import java.util.Vector;

class TypeRule {
	/**
	 * If we have operator X, and we have type Y on the left, and the Z on the right, 
	 * we can get the result type by looking at the type coercion rule table
	 */
	class TypeCoercionRule {
		DataType left;
		DataType right;
		String operator;
		DataType result;		
		TypeCoercionRule(String _operator,
				 DataType _left, 
				 DataType _right,				 
				 DataType _result) {
			left = _left;
			right = _right;
			operator = _operator;
			result = _result;
		}
		boolean match (String operator, 
			    DataType left,
			    DataType right)  {
			return (this.left==left &&
				this.right==right &&
				this.operator.equals(operator));
		}
		DataType getResult() {
			return result;
		}
	}
		
	Vector<TypeCoercionRule> info = new Vector<TypeCoercionRule>();

	void addTypeCoercionRule(String operator, 
				 DataType left,
				 DataType right,
				 DataType result, 
				 boolean symetric) {
		info.add(new TypeCoercionRule(operator, left, 
				      right, result));
		if (symetric && (left!=right)) {
			info.add(new TypeCoercionRule(operator, right, 
				      left, result));			
		}
	}

	/*add new type coercion rule*/
	void addTypeCoercionRule(String operator, 
				 DataType left,
				 DataType right,
				 DataType result
				 ) {
		addTypeCoercionRule(operator, left, right, result, false);
	}

	/*add new type coercion rule, 
	  shortcut to write addTypeCoercionRule(op, X, X, X);
	*/
	void addTypeCoercionRule(String operator, 
				 DataType left ) {
		addTypeCoercionRule(operator, left, left, left, false);
	}

	/*Boolean type coercion, always have result as integer	  
	*/
	void addTypeCoercionToBoolean(String operator, 
				      DataType left,
				      DataType right) {
		addTypeCoercionRule(operator, left, right, DataType.INTEGER, true);
	}

	/*Boolean type coercion, always have result as integer, 
	  this is a shortcut for addTypeCoercionToBoolean(op, X, X)
	*/
	void addTypeCoercionToBoolean(String operator, 
				      DataType left) {
		addTypeCoercionRule(operator, left, left, DataType.INTEGER, true);
	}

	/*implicit type conversion (coercion)*/
	DataType coerce(String operator, 
			       DataType left,
			       DataType right) throws InterpreterException {
		/*simple search, note, this is inefficient*/
		for (TypeCoercionRule t: info) {
			if (t.match(operator, left, right)) {
				return t.getResult();
			}
		}
		throw new InterpreterException("type mismatch "+left+" " + operator + " "+ right);
	}


	private TypeRule() {
		/*assignment, coerce to left type*/
		addTypeCoercionRule("=", DataType.STRING);
		addTypeCoercionRule("=", DataType.INTEGER);
		addTypeCoercionRule("=", DataType.FLOATING_POINT);
		addTypeCoercionRule("=", DataType.INTEGER, DataType.FLOATING_POINT, DataType.INTEGER);
		addTypeCoercionRule("=", DataType.FLOATING_POINT, DataType.INTEGER, DataType.FLOATING_POINT);		
		addTypeCoercionRule("+", DataType.STRING);

		String binary_operators [] = {"+", "-", "*", "/", "%"};
		
		for (String operator: binary_operators) {
			addTypeCoercionRule(operator, DataType.INTEGER);
			addTypeCoercionRule(operator, DataType.FLOATING_POINT);			
			/*operation with floating point results in floating point*/
			addTypeCoercionRule(operator, DataType.FLOATING_POINT, 
					    DataType.INTEGER, 
					    DataType.FLOATING_POINT, 
					    true
				);			
		}

		String compare_operators [] = {"<", ">", "<=", ">=", "==", "!="};
		for (String operator: compare_operators) {
			addTypeCoercionToBoolean(operator, DataType.INTEGER);
			addTypeCoercionToBoolean(operator, DataType.STRING);
			addTypeCoercionToBoolean(operator, DataType.FLOATING_POINT);
			addTypeCoercionToBoolean(operator, DataType.INTEGER, 
						 DataType.FLOATING_POINT);
		}

	}

	static TypeRule instance;

	static TypeRule getInstance() {
		if (instance==null) {
			instance = new TypeRule();
		}
		return instance;
	}

}
